@@ -44,6 +44,7 @@ int CrushWrapper::remove_item(CephContext *cct, int item)
 	  }	    
 	  was_bucket = t;
 	}
+	adjust_item_weight(cct, item, 0);
 	ldout(cct, 5) << "remove_device removing item " << item << " from bucket " << b->id << dendl;
 	crush_bucket_remove_item(b, item);
 	ret = 0;
@@ -59,14 +60,134 @@ int CrushWrapper::remove_item(CephContext *cct, int item)
     name_map.erase(item);
     have_rmaps = false;
     ret = 0;
-  }  
+  }
   
   return ret;
 }
 
-int CrushWrapper::insert_item(CephContext *cct, int item, float weight, string name,
-				map<string,string>& loc)  // typename -> bucketname
+bool CrushWrapper::check_item_loc(CephContext *cct, int item, map<string,string>& loc,
+				  int *weight)
 {
+  ldout(cct, 5) << "check_item_loc item " << item << " loc " << loc << dendl;
+
+  for (map<int,string>::const_iterator p = type_map.begin(); p != type_map.end(); p++) {
+    // ignore device
+    if (p->first == 0)
+      continue;
+
+    // ignore types that aren't specified in loc
+    if (loc.count(p->second) == 0) {
+      ldout(cct, 2) << "warning: did not specify location for '" << p->second << "' level (levels are "
+		    << type_map << ")" << dendl;
+      continue;
+    }
+
+    if (!name_exists(loc[p->second].c_str())) {
+      ldout(cct, 5) << "check_item_loc bucket " << loc[p->second] << " dne" << dendl;
+      return false;
+    }
+
+    int id = get_item_id(loc[p->second].c_str());
+    if (id >= 0) {
+      ldout(cct, 5) << "check_item_loc requested " << loc[p->second] << " for type " << p->second
+		    << " is a device, not bucket" << dendl;
+      return false;
+    }
+
+    crush_bucket *b = get_bucket(id);
+    assert(b);
+
+    // see if item exists in this bucket
+    for (unsigned j=0; j<b->size; j++) {
+      if (b->items[j] == item) {
+	ldout(cct, 2) << "check_item_loc " << item << " exists in bucket " << b->id << dendl;
+	if (weight)
+	  *weight = crush_get_bucket_item_weight(b, j);
+	return true;
+      }
+    }
+    return false;
+  }
+  
+  ldout(cct, 1) << "check_item_loc item " << item << " loc " << loc << dendl;
+  return false;
+}
+
+map<string, string> CrushWrapper::get_full_location(int id){
+
+  map<string,string> full_location;
+  pair<string, string> parent_coord;
+  parent_coord = get_immediate_parent(id);
+  int parent_id;
+
+  // read the type map and get the name of the type with the largest ID
+  int high_type = 0;
+  for (map<int, string>::iterator it = type_map.begin(); it != type_map.end(); it++){
+    if ( (*it).first > high_type )
+      high_type = (*it).first;
+  }
+
+  string high_type_name = type_map[high_type];
+
+  full_location[ parent_coord.first ] = parent_coord.second;
+  parent_id = get_item_id( (parent_coord.second).c_str() );
+
+
+  while (parent_coord.first != high_type_name) {
+    parent_coord = get_immediate_parent(parent_id);
+    full_location[ parent_coord.first ] = parent_coord.second;
+    if ( parent_coord.first != high_type_name ){
+      parent_id = get_item_id( (parent_coord.second).c_str() );
+    }
+  }
+
+  return full_location;
+}
+
+
+map<int, string> CrushWrapper::get_parent_hierarchy(int id)
+{
+  map<int,string> parent_hierarchy;
+  pair<string, string> parent_coord = get_immediate_parent(id);
+  int parent_id;
+
+  // get the integer type for id and create a counter from there
+  int type_counter = get_bucket_type(id);
+
+  // if we get a negative type then we can assume that we have an OSD
+  // change behavior in get_item_type FIXME
+  if (type_counter < 0)
+    type_counter = 0;
+
+  // read the type map and get the name of the type with the largest ID
+  int high_type = 0;
+  for (map<int, string>::iterator it = type_map.begin(); it != type_map.end(); it++){
+    if ( (*it).first > high_type )
+      high_type = (*it).first;
+  }
+
+  parent_id = get_item_id((parent_coord.second).c_str());
+
+  while (type_counter < high_type) {
+    type_counter++;
+    parent_hierarchy[ type_counter ] = parent_coord.first;
+
+    if (type_counter < high_type){
+      // get the coordinate information for the next parent
+      parent_coord = get_immediate_parent(parent_id);
+      parent_id = get_item_id(parent_coord.second.c_str());
+    }
+  }
+
+  return parent_hierarchy;
+}
+
+
+
+int CrushWrapper::insert_item(CephContext *cct, int item, float weight, string name,
+			      map<string,string>& loc)  // typename -> bucketname
+{
+
   ldout(cct, 5) << "insert_item item " << item << " weight " << weight
 		<< " name " << name << " loc " << loc << dendl;
 
@@ -81,29 +202,29 @@ int CrushWrapper::insert_item(CephContext *cct, int item, float weight, string n
   int cur = item;
 
   for (map<int,string>::iterator p = type_map.begin(); p != type_map.end(); p++) {
+    // ignore device type
     if (p->first == 0)
       continue;
 
+    // skip types that are unspecified
     if (loc.count(p->second) == 0) {
       ldout(cct, 2) << "warning: did not specify location for '" << p->second << "' level (levels are "
 		    << type_map << ")" << dendl;
       continue;
     }
 
-    int id = get_item_id(loc[p->second].c_str());
-    if (!id) {
-      // create the bucket
+    if (!name_exists(loc[p->second].c_str())) {
       ldout(cct, 5) << "insert_item creating bucket " << loc[p->second] << dendl;
       int empty = 0;
-      id = add_bucket(0, CRUSH_BUCKET_STRAW, CRUSH_HASH_DEFAULT, p->first, 1, &cur, &empty);
-      set_item_name(id, loc[p->second].c_str());
-      cur = id;
+      cur = add_bucket(0, CRUSH_BUCKET_STRAW, CRUSH_HASH_DEFAULT, p->first, 1, &cur, &empty);
+      set_item_name(cur, loc[p->second].c_str());
       continue;
     }
 
     // add to an existing bucket
+    int id = get_item_id(loc[p->second].c_str());
     if (!bucket_exists(id)) {
-      ldout(cct, 1) << "insert_item don't have bucket " << id << dendl;
+      ldout(cct, 1) << "insert_item doesn't have bucket " << id << dendl;
       return -EINVAL;
     }
 
@@ -119,10 +240,17 @@ int CrushWrapper::insert_item(CephContext *cct, int item, float weight, string n
     
     ldout(cct, 5) << "insert_item adding " << cur << " weight " << weight
 		  << " to bucket " << id << dendl;
-    crush_bucket_add_item(b, cur, 0);
+    int r = crush_bucket_add_item(b, cur, 0);
+    assert (!r);
 
     // now that we've added the (0-weighted) item and any parent buckets, adjust the weight.
     adjust_item_weightf(cct, item, weight);
+
+    if (item >= crush->max_devices) {
+      crush->max_devices = item + 1;
+      ldout(cct, 5) << "insert_item max_devices now " << crush->max_devices << dendl;
+    }
+
     return 0;
   }
 
@@ -130,6 +258,67 @@ int CrushWrapper::insert_item(CephContext *cct, int item, float weight, string n
   return -EINVAL;
 }
 
+int CrushWrapper::move_bucket(CephContext *cct, int id, map<string,string>& loc)
+{
+  // sorry this only works for buckets
+  if (id >= 0)
+    return -EINVAL;
+
+  if (!item_exists(id))
+    return -ENOENT;
+
+  // get the name of the bucket we are trying to move for later
+  string id_name = get_item_name(id);
+
+  // detach the bucket
+  int bucket_weight = detach_bucket(cct, id);
+
+  // un-set the device name so we can use add_item later
+  build_rmap(name_map, name_rmap);
+  name_map.erase(id);
+  name_rmap.erase(id_name);
+
+  // insert the bucket back into the hierarchy
+  return insert_item(cct, id, bucket_weight / (float)0x10000, id_name, loc);
+}
+
+
+int CrushWrapper::update_item(CephContext *cct, int item, float weight, string name,
+			      map<string,string>& loc)  // typename -> bucketname
+{
+  ldout(cct, 5) << "update_item item " << item << " weight " << weight
+		<< " name " << name << " loc " << loc << dendl;
+  int ret = 0;
+
+  // compare quantized (fixed-point integer) weights!  
+  int iweight = (int)(weight * (float)0x10000);
+  int old_iweight;
+  if (check_item_loc(cct, item, loc, &old_iweight)) {
+    ldout(cct, 5) << "update_item " << item << " already at " << loc << dendl;
+    if (old_iweight != iweight) {
+      ldout(cct, 5) << "update_item " << item << " adjusting weight "
+		    << ((float)old_iweight/(float)0x10000) << " -> " << weight << dendl;
+      adjust_item_weight(cct, item, iweight);
+      ret = 1;
+    }
+    if (get_item_name(item) != name) {
+      ldout(cct, 5) << "update_item setting " << item << " name to " << name << dendl;
+      set_item_name(item, name.c_str());
+      ret = 1;
+    }
+  } else {
+    if (item_exists(item)) {
+      remove_item(cct, item);
+    }
+    ldout(cct, 5) << "update_item adding " << item << " weight " << weight
+		  << " at " << loc << dendl;
+    ret = insert_item(cct, item, weight, name.c_str(), loc);
+    if (ret == 0)
+      ret = 1;  // changed
+  }
+  return ret;
+}
+
 int CrushWrapper::adjust_item_weight(CephContext *cct, int id, int weight)
 {
   ldout(cct, 5) << "adjust_item_weight " << id << " weight " << weight << dendl;
@@ -148,6 +337,45 @@ int CrushWrapper::adjust_item_weight(CephContext *cct, int id, int weight)
   return -ENOENT;
 }
 
+bool CrushWrapper::check_item_present(int id)
+{
+  bool found = false;
+
+  for (int bidx = 0; bidx < crush->max_buckets; bidx++) {
+    crush_bucket *b = crush->buckets[bidx];
+    if (b == 0)
+      continue;
+    for (unsigned i = 0; i < b->size; i++)
+      if (b->items[i] == id)
+	found = true;
+  }
+  return found;
+}
+
+
+pair<string,string> CrushWrapper::get_immediate_parent(int id)
+{
+  pair <string, string> loc;
+
+
+  for (int bidx = 0; bidx < crush->max_buckets; bidx++) {
+    crush_bucket *b = crush->buckets[bidx];
+    if (b == 0)
+      continue;
+    for (unsigned i = 0; i < b->size; i++)
+      if (b->items[i] == id){
+        string parent_id = name_map[b->id];
+        string parent_bucket_type = type_map[b->type];
+        loc = make_pair(parent_bucket_type, parent_id);
+      }
+  }
+
+
+  return loc;
+}
+
+
+
 void CrushWrapper::reweight(CephContext *cct)
 {
   set<int> roots;
@@ -157,7 +385,8 @@ void CrushWrapper::reweight(CephContext *cct)
       continue;
     crush_bucket *b = get_bucket(*p);
     ldout(cct, 5) << "reweight bucket " << *p << dendl;
-    crush_reweight_bucket(crush, b);
+    int r = crush_reweight_bucket(crush, b);
+    assert(r == 0);
   }
 }
 
@@ -237,6 +466,11 @@ void CrushWrapper::encode(bufferlist& bl, bool lean) const
   ::encode(type_map, bl);
   ::encode(name_map, bl);
   ::encode(rule_name_map, bl);
+
+  // tunables
+  ::encode(crush->choose_local_tries, bl);
+  ::encode(crush->choose_local_fallback_tries, bl);
+  ::encode(crush->choose_total_tries, bl);
 }
 
 void CrushWrapper::decode(bufferlist::iterator& blp)
@@ -284,6 +518,12 @@ void CrushWrapper::decode(bufferlist::iterator& blp)
     ::decode(rule_name_map, blp);
     build_rmaps();
 
+    // tunables
+    if (!blp.end()) {
+      ::decode(crush->choose_local_tries, blp);
+      ::decode(crush->choose_local_fallback_tries, blp);
+      ::decode(crush->choose_total_tries, blp);
+    }
     finalize();
   }
   catch (...) {
