@@ -12,12 +12,14 @@
  *
  */
 
+#include <limits.h>
+
 #include "IoCtxImpl.h"
 
 #include "librados/AioCompletionImpl.h"
 #include "librados/PoolAsyncCompletionImpl.h"
 #include "librados/RadosClient.h"
-
+#include "include/assert.h"
 
 #define dout_subsys ceph_subsys_rados
 #undef dout_prefix
@@ -646,13 +648,14 @@ int librados::IoCtxImpl::aio_operate_read(const object_t &oid,
 {
   Context *onack = new C_aio_Ack(c);
 
+  c->is_read = true;
   c->io = this;
   c->pbl = pbl;
 
   Mutex::Locker l(*lock);
   objecter->read(oid, oloc,
 		 *o, snap_seq, pbl, 0,
-		 onack, 0);
+		 onack, &c->objver);
   return 0;
 }
 
@@ -679,10 +682,13 @@ int librados::IoCtxImpl::aio_operate(const object_t& oid,
 int librados::IoCtxImpl::aio_read(const object_t oid, AioCompletionImpl *c,
 				  bufferlist *pbl, size_t len, uint64_t off)
 {
+  if (len > (size_t) INT_MAX)
+    return -EDOM;
 
   Context *onack = new C_aio_Ack(c);
   eversion_t ver;
 
+  c->is_read = true;
   c->io = this;
   c->pbl = pbl;
 
@@ -696,8 +702,12 @@ int librados::IoCtxImpl::aio_read(const object_t oid, AioCompletionImpl *c,
 int librados::IoCtxImpl::aio_read(const object_t oid, AioCompletionImpl *c,
 				  char *buf, size_t len, uint64_t off)
 {
+  if (len > (size_t) INT_MAX)
+    return -EDOM;
+
   Context *onack = new C_aio_Ack(c);
 
+  c->is_read = true;
   c->io = this;
   c->buf = buf;
   c->maxlen = len;
@@ -716,6 +726,8 @@ int librados::IoCtxImpl::aio_sparse_read(const object_t oid,
 					 bufferlist *data_bl, size_t len,
 					 uint64_t off)
 {
+  if (len > (size_t) INT_MAX)
+    return -EDOM;
 
   C_aio_sparse_read_Ack *onack = new C_aio_sparse_read_Ack(c);
   onack->m = m;
@@ -1010,6 +1022,7 @@ int librados::IoCtxImpl::aio_exec(const object_t& oid, AioCompletionImpl *c,
 {
   Context *onack = new C_aio_Ack(c);
 
+  c->is_read = true;
   c->io = this;
 
   Mutex::Locker l(*lock);
@@ -1024,6 +1037,9 @@ int librados::IoCtxImpl::aio_exec(const object_t& oid, AioCompletionImpl *c,
 int librados::IoCtxImpl::read(const object_t& oid,
 			      bufferlist& bl, size_t len, uint64_t off)
 {
+  if (len > (size_t) INT_MAX)
+    return -EDOM;
+
   Mutex mylock("IoCtxImpl::read::mylock");
   Cond cond;
   bool done;
@@ -1031,6 +1047,7 @@ int librados::IoCtxImpl::read(const object_t& oid,
   Context *onack = new C_SafeCond(&mylock, &cond, &done, &r);
   eversion_t ver;
 
+
   ::ObjectOperation op;
   ::ObjectOperation *pop = prepare_assert_ops(&op);
 
@@ -1097,6 +1114,9 @@ int librados::IoCtxImpl::sparse_read(const object_t& oid,
 				     bufferlist& data_bl, size_t len,
 				     uint64_t off)
 {
+  if (len > (size_t) INT_MAX)
+    return -EDOM;
+
   bufferlist bl;
 
   Mutex mylock("IoCtxImpl::read::mylock");
@@ -1489,6 +1509,8 @@ void librados::IoCtxImpl::C_aio_Ack::finish(int r)
   c->lock.Lock();
   c->rval = r;
   c->ack = true;
+  if (c->is_read)
+    c->safe = true;
   c->cond.Signal();
 
   if (c->buf && c->bl.length() > 0) {
@@ -1503,6 +1525,9 @@ void librados::IoCtxImpl::C_aio_Ack::finish(int r)
   if (c->callback_complete) {
     c->io->client->finisher.queue(new C_AioComplete(c));
   }
+  if (c->is_read && c->callback_safe) {
+    c->io->client->finisher.queue(new C_AioSafe(c));
+  }
 
   c->put_unlock();
 }
@@ -1575,8 +1600,10 @@ void librados::IoCtxImpl::C_NotifyComplete::notify(uint8_t opcode,
 						   uint64_t ver,
 						   bufferlist& bl)
 {
+  lock->Lock();
   *done = true;
   cond->Signal();
+  lock->Unlock();
 }
 
 /////////////////////////// WatchContext ///////////////////////////////
@@ -1594,13 +1621,16 @@ librados::WatchContext::~WatchContext()
   io_ctx_impl->put();
 }
 
-void librados::WatchContext::notify(uint8_t opcode,
+void librados::WatchContext::notify(Mutex *client_lock,
+                                    uint8_t opcode,
 				    uint64_t ver,
 				    uint64_t notify_id,
 				    bufferlist& payload)
 {
   ctx->notify(opcode, ver, payload);
   if (opcode != WATCH_NOTIFY_COMPLETE) {
+    client_lock->Lock();
     io_ctx_impl->_notify_ack(oid, notify_id, ver);
+    client_lock->Unlock();
   }
 }
