@@ -12,6 +12,8 @@
  * 
  */
 
+#include <sstream>
+
 #include "OSDMonitor.h"
 #include "Monitor.h"
 #include "MDSMonitor.h"
@@ -42,8 +44,7 @@
 #include "common/errno.h"
 
 #include "include/compat.h"
-
-#include <sstream>
+#include "include/assert.h"
 
 #define dout_subsys ceph_subsys_mon
 #undef dout_prefix
@@ -233,6 +234,9 @@ void OSDMonitor::on_active()
 
   if (thrash_map && thrash())
     propose_pending();
+
+  if (mon->is_leader())
+    mon->clog.info() << "osdmap " << osdmap << "\n"; 
 }
 
 void OSDMonitor::update_logger()
@@ -313,6 +317,7 @@ int OSDMonitor::reweight_by_utilization(int oload, std::string& out_str)
   oss << buf;
   std::string sep;
   oss << "overloaded osds: ";
+  bool changed = false;
   for (hash_map<int,osd_stat_t>::const_iterator p = pgm.osd_stat.begin();
        p != pgm.osd_stat.end();
        ++p) {
@@ -323,13 +328,15 @@ int OSDMonitor::reweight_by_utilization(int oload, std::string& out_str)
       // Assign a lower weight to overloaded OSDs. The current weight
       // is a factor to take into account the original weights,
       // to represent e.g. differing storage capacities
-      float weight = osdmap.get_weightf(p->first);
-      float new_weight = (average_util / util) * weight;
-      osdmap.set_weightf(p->first, new_weight);
+      unsigned weight = osdmap.get_weight(p->first);
+      unsigned new_weight = (unsigned)((average_util / util) * (float)weight);
+      pending_inc.new_weight[p->first] = new_weight;
       char buf[128];
       snprintf(buf, sizeof(buf), "%d [%04f -> %04f]", p->first,
-	       weight, new_weight);
+	       (float)weight / (float)0x10000,
+	       (float)new_weight / (float)0x10000);
       oss << buf << sep;
+      changed = true;
     }
   }
   if (sep.empty()) {
@@ -337,7 +344,7 @@ int OSDMonitor::reweight_by_utilization(int oload, std::string& out_str)
   }
   out_str = oss.str();
   dout(0) << "reweight_by_utilization: finished with " << out_str << dendl;
-  return 0;
+  return changed;
 }
 
 void OSDMonitor::create_pending()
@@ -389,7 +396,7 @@ void OSDMonitor::encode_pending(bufferlist &bl)
 
   // encode
   assert(paxos->get_version() + 1 == pending_inc.epoch);
-  ::encode(pending_inc, bl, -1);
+  ::encode(pending_inc, bl, CEPH_FEATURES_ALL);
 }
 
 
@@ -597,6 +604,10 @@ bool OSDMonitor::can_mark_up(int i)
   return true;
 }
 
+/**
+ * @note the parameter @p i apparently only exists here so we can output the
+ *	 osd's id on messages.
+ */
 bool OSDMonitor::can_mark_out(int i)
 {
   if (osdmap.test_flag(CEPH_OSDMAP_NOOUT)) {
@@ -1246,6 +1257,12 @@ void OSDMonitor::tick()
   // mark down osds out?
   utime_t now = ceph_clock_now(g_ceph_context);
 
+
+  /* can_mark_out() checks if we can mark osds as being out. The -1 has no
+   * influence at all. The decision is made based on the ratio of "in" osds,
+   * and the function returns false if this ratio is lower that the minimum
+   * ratio set by g_conf->mon_osd_min_in_ratio. So it's not really up to us.
+   */
   if (can_mark_out(-1)) {
     map<int,utime_t>::iterator i = down_pending_out.begin();
     while (i != down_pending_out.end()) {
@@ -1438,6 +1455,15 @@ bool OSDMonitor::preprocess_command(MMonCommand *m)
   bufferlist rdata;
   stringstream ss;
 
+  MonSession *session = m->get_session();
+  if (!session ||
+      (!session->caps.get_allow_all() &&
+       !session->caps.check_privileges(PAXOS_OSDMAP, MON_CAP_R) &&
+       !mon->_allowed_command(session, m->cmd))) {
+    mon->reply_command(m, -EACCES, "access denied", rdata, paxos->get_version());
+    return true;
+  }
+
   vector<const char*> args;
   for (unsigned i = 1; i < m->cmd.size(); i++)
     args.push_back(m->cmd[i].c_str());
@@ -1718,6 +1744,16 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
   stringstream ss;
   string rs;
   int err = -EINVAL;
+
+  MonSession *session = m->get_session();
+  if (!session ||
+      (!session->caps.get_allow_all() &&
+       !session->caps.check_privileges(PAXOS_OSDMAP, MON_CAP_W) &&
+       !mon->_allowed_command(session, m->cmd))) {
+    mon->reply_command(m, -EACCES, "access denied", paxos->get_version());
+    return true;
+  }
+
   if (m->cmd.size() > 1) {
     if ((m->cmd.size() == 2 && m->cmd[1] == "setcrushmap") ||
 	(m->cmd.size() == 3 && m->cmd[1] == "crush" && m->cmd[2] == "set")) {
@@ -1773,7 +1809,10 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	    break;
 	  string key(s, 0, pos-s);
 	  string value(pos+1);
-	  loc[key] = value;
+	  if (value.length())
+	    loc[key] = value;
+	  else
+	    loc.erase(key);
 	}
 
 	dout(0) << "adding/updating crush item id " << id << " name '" << name << "' weight " << weight
@@ -1805,6 +1844,57 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	}
       } while (false);
     }
+    else if (m->cmd.size() >= 5 && m->cmd[1] == "crush" && m->cmd[2] == "move") {
+      do {
+	// osd crush move <name> [<loc1> [<loc2> ...]]
+	string name = m->cmd[3];
+	map<string,string> loc;
+	for (unsigned i = 4; i < m->cmd.size(); ++i) {
+	  const char *s = m->cmd[i].c_str();
+	  const char *pos = strchr(s, '=');
+	  if (!pos)
+	    break;
+	  string key(s, 0, pos-s);
+	  string value(pos+1);
+	  if (value.length())
+	    loc[key] = value;
+	  else
+	    loc.erase(key);
+	}
+
+	dout(0) << "moving crush item name '" << name << "' to location " << loc << dendl;
+	bufferlist bl;
+	if (pending_inc.crush.length())
+	  bl = pending_inc.crush;
+	else
+	  osdmap.crush->encode(bl);
+
+	CrushWrapper newcrush;
+	bufferlist::iterator p = bl.begin();
+	newcrush.decode(p);
+
+	if (!newcrush.name_exists(name.c_str())) {
+	  err = -ENOENT;
+	  ss << "item " << name << " dne";
+	  break;
+	}
+	int id = newcrush.get_item_id(name.c_str());
+
+	if (!newcrush.check_item_loc(g_ceph_context, id, loc, (int *)NULL)) {
+	  err = newcrush.move_bucket(g_ceph_context, id, loc);
+	  if (err >= 0) {
+	    ss << "moved item id " << id << " name '" << name << "' to location " << loc << " in crush map";
+	    pending_inc.crush.clear();
+	    newcrush.encode(pending_inc.crush);
+	    getline(ss, rs);
+	    paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
+	    return true;
+	  }
+	} else {
+	    ss << "no need to move item id " << id << " name '" << name << "' to location " << loc << " in crush map";
+	}
+      } while (false);
+    }
     else if (m->cmd.size() > 3 && m->cmd[1] == "crush" && (m->cmd[2] == "rm" || m->cmd[2] == "remove")) {
       do {
 	// osd crush rm <id>
@@ -2243,6 +2333,26 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
 	  paxos->wait_for_commit(new Monitor::C_Command(mon, m, ret, rs, paxos->get_version()));
 	  return true;
 	}
+      } else if (m->cmd[2] == "rename" && m->cmd.size() == 5) {
+	int64_t pool = osdmap.lookup_pg_pool_name(m->cmd[3].c_str());
+	if (pool < 0) {
+	  ss << "unrecognized pool '" << m->cmd[3] << "'";
+	  err = -ENOENT;
+	} else if (osdmap.lookup_pg_pool_name(m->cmd[4].c_str()) >= 0) {
+	  ss << "pool '" << m->cmd[4] << "' already exists";
+	  err = -EEXIST;
+	} else {
+	  int ret = _prepare_rename_pool(pool, m->cmd[4]);
+	  if (ret == 0) {
+	    ss << "pool '" << m->cmd[3] << "' renamed to '" << m->cmd[4] << "'";
+	  } else {
+	    ss << "failed to rename pool '" << m->cmd[3] << "' to '" << m->cmd[4] << "': "
+	       << cpp_strerror(ret);
+	  }
+	  getline(ss, rs);
+	  paxos->wait_for_commit(new Monitor::C_Command(mon, m, ret, rs, paxos->get_version()));
+	  return true;
+	}
       } else if (m->cmd[2] == "set") {
 	if (m->cmd.size() != 6) {
 	  err = -EINVAL;
@@ -2369,11 +2479,16 @@ bool OSDMonitor::prepare_command(MMonCommand *m)
       }
       string out_str;
       err = reweight_by_utilization(oload, out_str);
-      if (err) {
-	ss << "FAILED to reweight-by-utilization: " << out_str;
+      if (err < 0) {
+	ss << "FAILED reweight-by-utilization: " << out_str;
       }
-      else {
+      else if (err == 0) {
+	ss << "no change: " << out_str;
+      } else {
 	ss << "SUCCESSFUL reweight-by-utilization: " << out_str;
+	getline(ss, rs);
+	paxos->wait_for_commit(new Monitor::C_Command(mon, m, 0, rs, paxos->get_version()));
+	return true;
       }
     }
     else if (m->cmd.size() == 3 && m->cmd[1] == "thrash") {
@@ -2403,7 +2518,7 @@ bool OSDMonitor::preprocess_pool_op(MPoolOp *m)
 
   if (!osdmap.get_pg_pool(m->pool)) {
     dout(10) << "attempt to delete non-existent pool id " << m->pool << dendl;
-    _pool_op_reply(m, -ENODATA, osdmap.get_epoch());
+    _pool_op_reply(m, -ENOENT, osdmap.get_epoch());
     return true;
   }
   
@@ -2585,6 +2700,25 @@ int OSDMonitor::_prepare_remove_pool(uint64_t pool)
   return 0;
 }
 
+int OSDMonitor::_prepare_rename_pool(uint64_t pool, string newname)
+{
+  dout(10) << "_prepare_rename_pool " << pool << dendl;
+  if (pending_inc.old_pools.count(pool)) {
+    dout(10) << "_prepare_rename_pool " << pool << " pending removal" << dendl;    
+    return -ENOENT;
+  }
+  for (map<int64_t,string>::iterator p = pending_inc.new_pool_names.begin();
+       p != pending_inc.new_pool_names.end();
+       ++p) {
+    if (p->second == newname) {
+      return -EEXIST;
+    }
+  }
+
+  pending_inc.new_pool_names[pool] = newname;
+  return 0;
+}
+
 bool OSDMonitor::prepare_pool_op_delete(MPoolOp *m)
 {
   int ret = _prepare_remove_pool(m->pool);
