@@ -23,6 +23,7 @@
 #include <sys/uio.h>
 #include <limits.h>
 #include <sys/user.h>
+#include <poll.h>
 
 #include "common/config.h"
 #include "global/global_init.h"
@@ -41,16 +42,14 @@
 
 #include "include/compat.h"
 
-#define DOUT_SUBSYS ms
+#define dout_subsys ceph_subsys_ms
 #undef dout_prefix
 #define dout_prefix _prefix(_dout, msgr)
 static ostream& _prefix(std::ostream *_dout, SimpleMessenger *msgr) {
-  return *_dout << "-- " << msgr->ms_addr << " ";
+  return *_dout << "-- " << msgr->get_myaddr() << " ";
 }
 
 
-#include "tcp.cc"
-
 /********************************************
  * Accepter
  */
@@ -146,20 +145,20 @@ int SimpleMessenger::Accepter::bind(entity_addr_t &bind_addr, int avoid_port1, i
     return -errno;
   }
   
-  msgr->ms_addr = bind_addr;
-  if (msgr->ms_addr != entity_addr_t())
+  msgr->my_inst.addr = bind_addr;
+  if (msgr->my_inst.addr != entity_addr_t())
     msgr->need_addr = false;
   else 
     msgr->need_addr = true;
 
-  if (msgr->ms_addr.get_port() == 0) {
-    msgr->ms_addr = listen_addr;
-    msgr->ms_addr.nonce = msgr->nonce;
+  if (msgr->my_inst.addr.get_port() == 0) {
+    msgr->my_inst.addr = listen_addr;
+    msgr->my_inst.addr.nonce = msgr->nonce;
   }
 
   msgr->init_local_pipe();
 
-  ldout(msgr->cct,1) << "accepter.bind ms_addr is " << msgr->ms_addr << " need_addr=" << msgr->need_addr << dendl;
+  ldout(msgr->cct,1) << "accepter.bind my_inst.addr is " << msgr->my_inst.addr << " need_addr=" << msgr->need_addr << dendl;
   msgr->did_bind = true;
   return 0;
 }
@@ -171,7 +170,7 @@ int SimpleMessenger::Accepter::rebind(int avoid_port)
   
   stop();
 
-  entity_addr_t addr = msgr->ms_addr;
+  entity_addr_t addr = msgr->my_inst.addr;
   int old_port = addr.get_port();
   addr.set_port(0);
 
@@ -236,7 +235,7 @@ void *SimpleMessenger::Accepter::entry()
       msgr->lock.Lock();
 
       if (!msgr->destination_stopped) {
-	Pipe *p = new Pipe(msgr, Pipe::STATE_ACCEPTING);
+	Pipe *p = new Pipe(msgr, Pipe::STATE_ACCEPTING, NULL);
 	p->sd = sd;
 	p->pipe_lock.Lock();
 	p->start_reader();
@@ -292,61 +291,65 @@ void SimpleMessenger::Accepter::stop()
  * end of the queue. If the queue is empty; it's removed.
  * The message is then delivered and the process starts again.
  */
-void SimpleMessenger::dispatch_entry()
+void SimpleMessenger::DispatchQueue::entry()
 {
-  dispatch_queue.lock.Lock();
-  while (!dispatch_queue.stop) {
-    while (!dispatch_queue.queued_pipes.empty() && !dispatch_queue.stop) {
+  lock.Lock();
+  while (!stop) {
+    while (!queued_pipes.empty() && !stop) {
       //get highest-priority pipe
-      map<int, xlist<Pipe *>* >::reverse_iterator high_iter =
-	dispatch_queue.queued_pipes.rbegin();
+      map<int, xlist<IncomingQueue *>* >::reverse_iterator high_iter =
+	queued_pipes.rbegin();
       int priority = high_iter->first;
-      xlist<Pipe *> *pipe_list = high_iter->second;
+      xlist<IncomingQueue *> *qlist = high_iter->second;
 
-      Pipe *pipe = pipe_list->front();
+      IncomingQueue *inq = qlist->front();
       //move pipe to back of line -- or just take off if no more messages
-      pipe->pipe_lock.Lock();
-      list<Message *>& m_queue = pipe->in_q[priority];
+      inq->lock.Lock();
+      list<Message *>& m_queue = inq->in_q[priority];
       Message *m = m_queue.front();
       m_queue.pop_front();
 
       if (m_queue.empty()) {
-	pipe_list->pop_front();  // pipe is done
-	if (pipe_list->empty()) {
-	  delete pipe_list;
-	  dispatch_queue.queued_pipes.erase(priority);
+	qlist->pop_front();  // pipe is done
+	if (qlist->empty()) {
+	  delete qlist;
+	  queued_pipes.erase(priority);
 	}
+	inq->in_q.erase(priority);
+	ldout(cct,20) << "dispatch_entry inq " << inq << " pipe " << inq->pipe << " dequeued " << m
+		      << ", dequeued queue" << dendl;
       } else {
-	pipe_list->push_back(pipe->queue_items[priority]);  // move to end of list
+	ldout(cct,20) << "dispatch_entry inq " << inq << " pipe " << inq->pipe << " dequeued " << m
+		      << ", moved to end of list" << dendl;
+	qlist->push_back(inq->queue_items[priority]);  // move to end of list
       }
-      ldout(cct,20) << "dispatch_entry pipe " << pipe << " dequeued " << m << dendl;
-      dispatch_queue.lock.Unlock(); //done with the pipe queue for a while
+      lock.Unlock(); //done with the pipe queue for a while
 
-      pipe->in_qlen--;
-      dispatch_queue.qlen.dec();
+      inq->in_qlen--;
+      qlen.dec();
 
-      pipe->pipe_lock.Unlock(); // done with the pipe's message queue now
-       {
+      inq->lock.Unlock(); // done with the pipe's message queue now
+      {
 	if ((long)m == DispatchQueue::D_BAD_REMOTE_RESET) {
-	  dispatch_queue.lock.Lock();
-	  Connection *con = dispatch_queue.remote_reset_q.front();
-	  dispatch_queue.remote_reset_q.pop_front();
-	  dispatch_queue.lock.Unlock();
-	  ms_deliver_handle_remote_reset(con);
+	  lock.Lock();
+	  Connection *con = remote_reset_q.front();
+	  remote_reset_q.pop_front();
+	  lock.Unlock();
+	  msgr->ms_deliver_handle_remote_reset(con);
 	  con->put();
 	} else if ((long)m == DispatchQueue::D_CONNECT) {
-	  dispatch_queue.lock.Lock();
-	  Connection *con = dispatch_queue.connect_q.front();
-	  dispatch_queue.connect_q.pop_front();
-	  dispatch_queue.lock.Unlock();
-	  ms_deliver_handle_connect(con);
+	  lock.Lock();
+	  Connection *con = connect_q.front();
+	  connect_q.pop_front();
+	  lock.Unlock();
+	  msgr->ms_deliver_handle_connect(con);
 	  con->put();
 	} else if ((long)m == DispatchQueue::D_BAD_RESET) {
-	  dispatch_queue.lock.Lock();
-	  Connection *con = dispatch_queue.reset_q.front();
-	  dispatch_queue.reset_q.pop_front();
-	  dispatch_queue.lock.Unlock();
-	  ms_deliver_handle_reset(con);
+	  lock.Lock();
+	  Connection *con = reset_q.front();
+	  reset_q.pop_front();
+	  lock.Unlock();
+	  msgr->ms_deliver_handle_reset(con);
 	  con->put();
 	} else {
 	  uint64_t msize = m->get_dispatch_throttle_size();
@@ -361,19 +364,24 @@ void SimpleMessenger::dispatch_entry()
 		  << " " << m->get_footer().data_crc << ")"
 		  << " " << m << " con " << m->get_connection()
 		  << dendl;
-	  ms_deliver_dispatch(m);
+	  msgr->ms_deliver_dispatch(m);
 
-	  dispatch_throttle_release(msize);
+	  msgr->dispatch_throttle_release(msize);
 
 	  ldout(cct,20) << "done calling dispatch on " << m << dendl;
 	}
       }
-      dispatch_queue.lock.Lock();
+      lock.Lock();
     }
-    if (!dispatch_queue.stop)
-      dispatch_queue.cond.Wait(dispatch_queue.lock); //wait for something to be put on queue
+    if (!stop)
+      cond.Wait(lock); //wait for something to be put on queue
   }
-  dispatch_queue.lock.Unlock();
+  lock.Unlock();
+}
+
+void SimpleMessenger::dispatch_entry()
+{
+  dispatch_queue.entry();
 
   //tell everything else it's time to stop
   lock.Lock();
@@ -411,99 +419,64 @@ int SimpleMessenger::shutdown()
   return 0;
 }
 
-void SimpleMessenger::suicide()
-{
-  ldout(cct,10) << "suicide " << get_myaddr() << dendl;
-  shutdown();
-  // hmm, or exit(0)?
-}
-
-void SimpleMessenger::prepare_dest(const entity_inst_t& inst)
-{
-  lock.Lock();
-  {
-    if (rank_pipe.count(inst.addr) == 0)
-      connect_rank(inst.addr, inst.name.type());
-  }
-  lock.Unlock();
-}
-
-int SimpleMessenger::send_message(Message *m, const entity_inst_t& dest)
+int SimpleMessenger::_send_message(Message *m, const entity_inst_t& dest,
+                                   bool lazy)
 {
   // set envelope
   m->get_header().src = get_myname();
 
   if (!m->get_priority()) m->set_priority(get_default_send_priority());
  
-  ldout(cct,1) << "--> " << dest.name << " " << dest.addr
-          << " -- " << *m
+  ldout(cct,1) << (lazy ? "lazy " : "") <<"--> " << dest.name << " "
+          << dest.addr << " -- " << *m
     	  << " -- ?+" << m->get_data().length()
 	  << " " << m 
 	  << dendl;
 
-  submit_message(m, dest.addr, dest.name.type(), false);
+  if (dest.addr == entity_addr_t()) {
+    ldout(cct,0) << (lazy ? "lazy_" : "") << "send_message message " << *m
+                 << " with empty dest " << dest.addr << dendl;
+    m->put();
+    return -EINVAL;
+  }
+
+  lock.Lock();
+  Pipe *pipe = rank_pipe.count(dest.addr) ? rank_pipe[ dest.addr ] : NULL;
+  submit_message(m, (pipe ? pipe->connection_state : NULL),
+                 dest.addr, dest.name.type(), lazy);
+  lock.Unlock();
   return 0;
 }
 
-int SimpleMessenger::send_message(Message *m, Connection *con)
+int SimpleMessenger::_send_message(Message *m, Connection *con, bool lazy)
 {
   //set envelope
   m->get_header().src = get_myname();
 
   if (!m->get_priority()) m->set_priority(get_default_send_priority());
 
-  SimpleMessenger::Pipe *pipe = (SimpleMessenger::Pipe *)con->get_pipe();
-  if (pipe) {
-    ldout(cct,1) << "--> " << con->get_peer_addr() << " -- " << *m
-            << " -- ?+" << m->get_data().length()
-            << " " << m << " con " << con
-            << dendl;
+  ldout(cct,1) << (lazy ? "lazy " : "") << "--> " << con->get_peer_addr()
+      << " -- " << *m
+      << " -- ?+" << m->get_data().length()
+      << " " << m << " con " << con
+      << dendl;
 
-    submit_message(m, pipe);
-    pipe->put();
-  } else {
-    ldout(cct,0) << "send_message dropped message " << *m << " because of no pipe on con " << con
-	    << dendl;
-    // else we raced with reaper()
-    m->put();
-  }
+  lock.Lock();
+  submit_message(m, con, con->get_peer_addr(), con->get_peer_type(), lazy);
+  lock.Unlock();
   return 0;
 }
 
-int SimpleMessenger::lazy_send_message(Message *m, const entity_inst_t& dest)
-{
-  // set envelope
-  m->get_header().src = get_myname();
-
-  if (!m->get_priority()) m->set_priority(get_default_send_priority());
- 
-  ldout(cct,1) << "lazy "
-	  << " --> " << dest.name << " " << dest.addr
-          << " -- " << *m
-    	  << " -- ?+" << m->get_data().length()
-          << " " << m
-          << dendl;
-
-  submit_message(m, dest.addr, dest.name.type(), true);
-  return 0;
-}
-
-entity_addr_t SimpleMessenger::get_myaddr()
-{
-  entity_addr_t a = ms_addr;
-  return a;  
-}
-
 /**
- * If ms_addr doesn't have an IP set, this function
+ * If my_inst.addr doesn't have an IP set, this function
  * will fill it in from the passed addr. Otherwise it does nothing and returns.
  */
-void SimpleMessenger::set_ip(entity_addr_t &addr)
+void SimpleMessenger::set_addr_unknowns(entity_addr_t &addr)
 {
-  if (ms_addr.is_blank_ip()) {
-    int port = ms_addr.get_port();
-    ms_addr.addr = addr.addr;
-    ms_addr.set_port(port);
+  if (my_inst.addr.is_blank_ip()) {
+    int port = my_inst.addr.get_port();
+    my_inst.addr.addr = addr.addr;
+    my_inst.addr.set_port(port);
   }
 }
 
@@ -533,6 +506,9 @@ int SimpleMessenger::get_proto_version(int peer_type, bool connect)
 }
 
 
+
+
+
 /**************************************
  * Pipe
  */
@@ -540,7 +516,7 @@ int SimpleMessenger::get_proto_version(int peer_type, bool connect)
 #undef dout_prefix
 #define dout_prefix _pipe_prefix(_dout)
 ostream& SimpleMessenger::Pipe::_pipe_prefix(std::ostream *_dout) {
-  return *_dout << "-- " << msgr->ms_addr << " >> " << peer_addr << " pipe(" << this
+  return *_dout << "-- " << msgr->my_inst.addr << " >> " << peer_addr << " pipe(" << this
 		<< " sd=" << sd
 		<< " pgs=" << peer_global_seq
 		<< " cs=" << connect_seq
@@ -551,61 +527,7 @@ ostream& SimpleMessenger::Pipe::_pipe_prefix(std::ostream *_dout) {
 void SimpleMessenger::Pipe::queue_received(Message *m, int priority)
 {
   assert(pipe_lock.is_locked());
-  
-  list<Message *>& queue = in_q[priority];
-  
-  if (halt_delivery)
-    goto halt;
-  
-  if (queue.empty()) {
-    // queue pipe AND message under pipe AND dispatch_queue locks.
-    pipe_lock.Unlock();
-    msgr->dispatch_queue.lock.Lock();
-    pipe_lock.Lock();
-
-    if (halt_delivery) {
-      msgr->dispatch_queue.lock.Unlock();
-      goto halt;
-    }
-
-    if (queue.empty()) {
-      ldout(msgr->cct,20) << "queue_received queuing pipe" << dendl;
-      if (!queue_items.count(priority)) 
-	queue_items[priority] = new xlist<Pipe *>::item(this);
-      if (msgr->dispatch_queue.queued_pipes.empty())
-	msgr->dispatch_queue.cond.Signal();
-
-      map<int, xlist<Pipe*>*>::iterator p = msgr->dispatch_queue.queued_pipes.find(priority);
-      xlist<Pipe*> *pipe_list;
-      if (p != msgr->dispatch_queue.queued_pipes.end())
-	pipe_list = p->second;
-      else {
-	pipe_list = new xlist<Pipe*>;
-	msgr->dispatch_queue.queued_pipes[priority] = pipe_list;
-      }
-      pipe_list->push_back(queue_items[priority]);
-    }
-
-    queue.push_back(m);
-
-    msgr->dispatch_queue.lock.Unlock();
-  } else {
-    // just queue message under pipe lock.
-    queue.push_back(m);
-  }
-  
-  // increment queue length counters
-  in_qlen++;
-  msgr->dispatch_queue.qlen.inc();
-  
-  return;
-  
- halt:
-  // don't want to put local-delivery signals
-  if (m>(void *)DispatchQueue::D_NUM_CODES) {
-    msgr->dispatch_throttle_release(m->get_dispatch_throttle_size());
-    m->put();
-  }
+  in_q->queue(m, priority, &msgr->dispatch_queue);
 }
 
 
@@ -627,7 +549,7 @@ int SimpleMessenger::Pipe::accept()
 
   // and my addr
   bufferlist addrs;
-  ::encode(msgr->ms_addr, addrs);
+  ::encode(msgr->my_inst.addr, addrs);
 
   // and peer's socket addr (they might not know their ip)
   entity_addr_t socket_addr;
@@ -698,6 +620,7 @@ int SimpleMessenger::Pipe::accept()
   bufferlist authorizer, authorizer_reply;
   bool authorizer_valid;
   uint64_t feat_missing;
+  bool replaced = false;
 
   // this should roughly mirror pseudocode at
   //  http://ceph.newdream.net/wiki/Messaging_protocol
@@ -831,11 +754,12 @@ int SimpleMessenger::Pipe::accept()
 
       if (connect.connect_seq == existing->connect_seq) {
 	// connection race?
-	if (peer_addr < msgr->ms_addr ||
-	    existing->policy.server) {
+	if (peer_addr < msgr->my_inst.addr ||
+	    existing->policy.server ||
+	    existing->state == STATE_STANDBY) {
 	  // incoming wins
 	  ldout(msgr->cct,10) << "accept connection race, existing " << existing << ".cseq " << existing->connect_seq
-		   << " == " << connect.connect_seq << ", or we are server, replacing my attempt" << dendl;
+		   << " == " << connect.connect_seq << ", or STANDBY, or we are server, replacing my attempt" << dendl;
 	  if (!(existing->state == STATE_CONNECTING ||
 		existing->state == STATE_STANDBY ||
 		existing->state == STATE_WAIT))
@@ -851,7 +775,7 @@ int SimpleMessenger::Pipe::accept()
 	  // our existing outgoing wins
 	  ldout(msgr->cct,10) << "accept connection race, existing " << existing << ".cseq " << existing->connect_seq
 		   << " == " << connect.connect_seq << ", sending WAIT" << dendl;
-	  assert(peer_addr > msgr->ms_addr);
+	  assert(peer_addr > msgr->my_inst.addr);
 	  if (!(existing->state == STATE_CONNECTING ||
 		existing->state == STATE_OPEN))
 	    lderr(msgr->cct) << "accept race bad state, would send wait, existing=" << existing->state
@@ -859,7 +783,7 @@ int SimpleMessenger::Pipe::accept()
 			     << " == " << connect.connect_seq
 			     << dendl;
 	  assert(existing->state == STATE_CONNECTING ||
-		 existing->state == STATE_OPEN); // this will win
+		 existing->state == STATE_OPEN);
 	  reply.tag = CEPH_MSGR_TAG_WAIT;
 	  existing->pipe_lock.Unlock();
 	  msgr->lock.Unlock();
@@ -919,6 +843,7 @@ int SimpleMessenger::Pipe::accept()
   ldout(msgr->cct,10) << "accept replacing " << existing << dendl;
   existing->stop();
   existing->unregister_pipe();
+  replaced = true;
     
   if (!existing->policy.lossy) { /* if we're lossy, we can lose messages and
                                     should let the daemon handle it itself.
@@ -928,11 +853,20 @@ int SimpleMessenger::Pipe::accept()
     // do not clear existing->connection_state, since read_message and write_message both
     // dereference it without pipe_lock.
 
-    // steal queue and out_seq
-    existing->requeue_sent();
-    out_seq = existing->out_seq;
+    // steal incoming queue
     in_seq = existing->in_seq;
     in_seq_acked = in_seq;
+    delete in_q;
+    in_q = existing->in_q;
+    in_q->lock.Lock();
+    in_q->pipe = this;
+    in_q->restart_queue();
+    in_q->lock.Unlock();
+    existing->in_q = new IncomingQueue(msgr->cct, existing);
+
+    // steal outgoing queue and out_seq
+    existing->requeue_sent();
+    out_seq = existing->out_seq;
     ldout(msgr->cct,10) << "accept re-queuing on out_seq " << out_seq << " in_seq " << in_seq << dendl;
     for (map<int, list<Message*> >::iterator p = existing->out_q.begin();
          p != existing->out_q.end();
@@ -985,7 +919,7 @@ int SimpleMessenger::Pipe::accept()
       ldout(msgr->cct,2) << "accept write error on in_seq" << dendl;
       goto fail_unlocked;
     }
-    if(tcp_read(msgr->cct, sd, (char*)&newly_acked_seq, sizeof(newly_acked_seq)) < 0) {
+    if (tcp_read(msgr->cct, sd, (char*)&newly_acked_seq, sizeof(newly_acked_seq)) < 0) {
       ldout(msgr->cct,2) << "accept read error on newly_acked_seq" << dendl;
       goto fail_unlocked;
     }
@@ -1003,14 +937,16 @@ int SimpleMessenger::Pipe::accept()
 
  fail_unlocked:
   pipe_lock.Lock();
-  {
+  if (state != STATE_CLOSED) {
     bool queued = is_queued();
-    if (queued && state != STATE_CLOSED)
+    if (queued)
       state = STATE_CONNECTING;
+    else if (replaced)
+      state = STATE_STANDBY;
     else
       state = STATE_CLOSED;
     fault();
-    if (queued)
+    if (queued || replaced)
       start_writer();
   }
   pipe_lock.Unlock();
@@ -1100,7 +1036,7 @@ int SimpleMessenger::Pipe::connect()
   msg.msg_iov = msgvec;
   msg.msg_iovlen = 1;
   msglen = msgvec[0].iov_len;
-  if (do_sendmsg(sd, &msg, msglen)) {
+  if (do_sendmsg(&msg, msglen)) {
     ldout(msgr->cct,2) << "connect couldn't write my banner, " << strerror_r(errno, buf, sizeof(buf)) << dendl;
     goto fail;
   }
@@ -1140,7 +1076,7 @@ int SimpleMessenger::Pipe::connect()
   if (msgr->need_addr)
     msgr->learned_addr(peer_addr_for_me);
 
-  ::encode(msgr->ms_addr, myaddrbl);
+  ::encode(msgr->my_inst.addr, myaddrbl);
 
   memset(&msg, 0, sizeof(msg));
   msgvec[0].iov_base = myaddrbl.c_str();
@@ -1148,11 +1084,11 @@ int SimpleMessenger::Pipe::connect()
   msg.msg_iov = msgvec;
   msg.msg_iovlen = 1;
   msglen = msgvec[0].iov_len;
-  if (do_sendmsg(sd, &msg, msglen)) {
+  if (do_sendmsg(&msg, msglen)) {
     ldout(msgr->cct,2) << "connect couldn't write my addr, " << strerror_r(errno, buf, sizeof(buf)) << dendl;
     goto fail;
   }
-  ldout(msgr->cct,10) << "connect sent my addr " << msgr->ms_addr << dendl;
+  ldout(msgr->cct,10) << "connect sent my addr " << msgr->my_inst.addr << dendl;
 
 
   while (1) {
@@ -1189,7 +1125,7 @@ int SimpleMessenger::Pipe::connect()
 
     ldout(msgr->cct,10) << "connect sending gseq=" << gseq << " cseq=" << cseq
 	     << " proto=" << connect.protocol_version << dendl;
-    if (do_sendmsg(sd, &msg, msglen)) {
+    if (do_sendmsg(&msg, msglen)) {
       ldout(msgr->cct,2) << "connect couldn't write gseq, cseq, " << strerror_r(errno, buf, sizeof(buf)) << dendl;
       goto fail;
     }
@@ -1259,7 +1195,7 @@ int SimpleMessenger::Pipe::connect()
     if (reply.tag == CEPH_MSGR_TAG_RESETSESSION) {
       ldout(msgr->cct,0) << "connect got RESETSESSION" << dendl;
       was_session_reset();
-      halt_delivery = false;
+      in_q->restart_queue();
       cseq = 0;
       pipe_lock.Unlock();
       continue;
@@ -1310,9 +1246,7 @@ int SimpleMessenger::Pipe::connect()
 
       // hooray!
       peer_global_seq = reply.global_seq;
-      if (!disposable) {
-	policy.lossy = reply.flags & CEPH_MSG_CONNECT_LOSSY;
-      }
+      policy.lossy = reply.flags & CEPH_MSG_CONNECT_LOSSY;
       state = STATE_OPEN;
       connect_seq = cseq + 1;
       assert(connect_seq == reply.connect_seq);
@@ -1402,39 +1336,9 @@ void SimpleMessenger::Pipe::discard_queue()
 {
   ldout(msgr->cct,10) << "discard_queue" << dendl;
 
-  halt_delivery = true;
-
-  // dequeue pipe
-  DispatchQueue& q = msgr->dispatch_queue;
-  pipe_lock.Unlock();
-  q.lock.Lock();
-  pipe_lock.Lock();
-  for (map<int, xlist<Pipe *>::item* >::iterator i = queue_items.begin();
-       i != queue_items.end();
-       ++i) {
-    xlist<Pipe *>* list_on;
-    if ((list_on = i->second->get_list())) { //if in round-robin
-      i->second->remove_myself(); //take off
-      if (list_on->empty()) { //if round-robin queue is empty
-	delete list_on;
-	q.queued_pipes.erase(i->first); //remove from map
-      }
-    }
-  }
-
-  // clear queue_items
-  while (!queue_items.empty()) {
-    delete queue_items.begin()->second;
-    queue_items.erase(queue_items.begin());
-  }
-
-  q.lock.Unlock();
-
+  in_q->discard_queue(msgr, &msgr->dispatch_queue);
   ldout(msgr->cct,20) << " dequeued pipe " << dendl;
 
-  // adjust qlen
-  q.qlen.sub(in_qlen);
-
   for (list<Message*>::iterator p = sent.begin(); p != sent.end(); p++) {
     if (*p < (void *) DispatchQueue::D_NUM_CODES) {
       continue; // skip non-Message dispatch codes
@@ -1452,17 +1356,6 @@ void SimpleMessenger::Pipe::discard_queue()
       (*r)->put();
     }
   out_q.clear();
-  for (map<int,list<Message*> >::iterator p = in_q.begin(); p != in_q.end(); p++)
-    for (list<Message*>::iterator r = p->second.begin(); r != p->second.end(); r++) {
-      if (*r < (void *) DispatchQueue::D_NUM_CODES) {
-        continue; // skip non-Message dispatch codes
-      }
-      msgr->dispatch_throttle_release((*r)->get_dispatch_throttle_size());
-      ldout(msgr->cct,20) << "  discard " << *r << dendl;
-      (*r)->put();
-    }
-  in_q.clear();
-  in_qlen = 0;
 }
 
 
@@ -1809,10 +1702,8 @@ void SimpleMessenger::Pipe::writer()
     }
     
     if (sent.empty() && close_on_empty) {
-      // this is slightly hacky
       ldout(msgr->cct,10) << "writer out and sent queues empty, closing" << dendl;
-      policy.lossy = true;
-      fault();
+      stop();
       continue;
     }
 
@@ -1930,10 +1821,7 @@ int SimpleMessenger::Pipe::read_message(Message **pm)
     waited_on_throttle |= msgr->dispatch_throttler.get(message_size);
   }
 
-  utime_t throttle_wait;
-  if (waited_on_throttle) {
-    throttle_wait = ceph_clock_now(msgr->cct) - recv_stamp;
-  }
+  utime_t throttle_stamp = ceph_clock_now(msgr->cct);
 
   // read front
   front_len = header.front_len;
@@ -2041,7 +1929,8 @@ int SimpleMessenger::Pipe::read_message(Message **pm)
   message->set_dispatch_throttle_size(message_size);
 
   message->set_recv_stamp(recv_stamp);
-  message->set_throttle_wait(throttle_wait);
+  message->set_throttle_stamp(throttle_stamp);
+  message->set_recv_complete_stamp(ceph_clock_now(msgr->cct));
 
   *pm = message;
   return 0;
@@ -2061,7 +1950,7 @@ int SimpleMessenger::Pipe::read_message(Message **pm)
   return ret;
 }
 
-int SimpleMessenger::Pipe::do_sendmsg(int sd, struct msghdr *msg, int len, bool more)
+int SimpleMessenger::Pipe::do_sendmsg(struct msghdr *msg, int len, bool more)
 {
   char buf[80];
 
@@ -2086,40 +1975,6 @@ int SimpleMessenger::Pipe::do_sendmsg(int sd, struct msghdr *msg, int len, bool
       return -1; // close enough
     }
 
-    if (0) {
-      // hex dump
-      struct iovec *v = msg->msg_iov;
-      size_t left = r;
-      size_t vpos = 0;
-      ldout(msgr->cct,0) << "do_sendmsg wrote " << r << " bytes, hexdump:\n";
-      int pos = 0;
-      int col = 0;
-      char buf[20];
-      while (left > 0) {
-	if (col == 0) {
-	  snprintf(buf, sizeof(buf), "%05x : ", pos);
-	  *_dout << buf;
-	}
-	snprintf(buf, sizeof(buf), " %02x", ((unsigned char*)v->iov_base)[vpos]);
-	*_dout << buf;
-	left--;
-	if (!left)
-	  break;
-	vpos++;
-	pos++;
-	if (vpos == v->iov_len) {
-	  v++;
-	  vpos = 0;
-	}	  
-	col++;
-	if (col == 16) {
-	  *_dout << "\n";
-	  col = 0;
-	}
-      }
-      *_dout << dendl;
-    }
-
     len -= r;
     if (len == 0) break;
     
@@ -2163,7 +2018,7 @@ int SimpleMessenger::Pipe::write_ack(uint64_t seq)
   msg.msg_iov = msgvec;
   msg.msg_iovlen = 2;
   
-  if (do_sendmsg(sd, &msg, 1 + sizeof(s), true) < 0) 
+  if (do_sendmsg(&msg, 1 + sizeof(s), true) < 0)
     return -1;	
   return 0;
 }
@@ -2182,7 +2037,7 @@ int SimpleMessenger::Pipe::write_keepalive()
   msg.msg_iov = msgvec;
   msg.msg_iovlen = 1;
   
-  if (do_sendmsg(sd, &msg, 1) < 0) 
+  if (do_sendmsg(&msg, 1) < 0)
     return -1;	
   return 0;
 }
@@ -2262,7 +2117,7 @@ int SimpleMessenger::Pipe::write_message(Message *m)
 	     << dendl;
     
     if (msg.msg_iovlen >= IOV_MAX-2) {
-      if (do_sendmsg(sd, &msg, msglen, true)) 
+      if (do_sendmsg(&msg, msglen, true))
 	goto fail;
       
       // and restart the iov
@@ -2296,7 +2151,7 @@ int SimpleMessenger::Pipe::write_message(Message *m)
   msg.msg_iovlen++;
 
   // send
-  if (do_sendmsg(sd, &msg, msglen)) 
+  if (do_sendmsg(&msg, msglen))
     goto fail;
 
   ret = 0;
@@ -2311,6 +2166,119 @@ int SimpleMessenger::Pipe::write_message(Message *m)
 }
 
 
+/**************************************
+ * IncomingQueue
+ */
+
+#undef dout_prefix
+#define dout_prefix pipe->_pipe_prefix(_dout) << "incomingqueue."
+
+void SimpleMessenger::IncomingQueue::queue(Message *m, int priority, DispatchQueue *dq)
+{
+  Mutex::Locker l(lock);
+  ldout(cct,20) << "queue " << m << " prio " << priority << dendl;
+  if (in_q.count(priority) == 0) {
+    // queue inq AND message under inq AND dispatch_queue locks.
+    lock.Unlock();
+    dq->lock.Lock();
+    lock.Lock();
+
+    if (halt) {
+      dq->lock.Unlock();
+      goto halt;
+    }
+
+    list<Message *>& queue = in_q[priority];
+    if (queue.empty()) {
+      ldout(cct,20) << "queue " << m << " under newly queued queue" << dendl;
+      if (!queue_items.count(priority)) 
+	queue_items[priority] = new xlist<IncomingQueue *>::item(this);
+      if (dq->queued_pipes.empty())
+	dq->cond.Signal();
+
+      map<int, xlist<IncomingQueue*>*>::iterator p = dq->queued_pipes.find(priority);
+      xlist<IncomingQueue*> *qlist;
+      if (p != dq->queued_pipes.end())
+	qlist = p->second;
+      else {
+	qlist = new xlist<IncomingQueue*>;
+	dq->queued_pipes[priority] = qlist;
+      }
+      qlist->push_back(queue_items[priority]);
+    }
+
+    queue.push_back(m);
+
+    dq->lock.Unlock();
+  } else {
+    ldout(cct,20) << "queue " << m << " under existing queue" << dendl;
+    // just queue message under our lock.
+    list<Message *>& queue = in_q[priority];
+    queue.push_back(m);
+  }
+  
+  // increment queue length counters
+  in_qlen++;
+  dq->qlen.inc();
+  return;
+
+ halt:
+  ldout(cct, 20) << "queue " << m << " halt, discarding" << dendl;
+  // don't want to put local-delivery signals
+  if (m>(void *)DispatchQueue::D_NUM_CODES) {
+    pipe->msgr->dispatch_throttle_release(m->get_dispatch_throttle_size());
+    m->put();
+  }
+}
+
+void SimpleMessenger::IncomingQueue::discard_queue(SimpleMessenger *msgr, DispatchQueue *dq)
+{
+  halt = true;
+
+  // dequeue ourselves
+  dq->lock.Lock();
+  for (map<int, xlist<IncomingQueue *>::item* >::iterator i = queue_items.begin();
+       i != queue_items.end();
+       ++i) {
+    xlist<IncomingQueue *>* list_on;
+    if ((list_on = i->second->get_list())) { //if in round-robin
+      i->second->remove_myself(); //take off
+      if (list_on->empty()) { //if round-robin queue is empty
+	delete list_on;
+	dq->queued_pipes.erase(i->first); //remove from map
+      }
+    }
+  }
+  dq->lock.Unlock();
+
+  while (!queue_items.empty()) {
+    delete queue_items.begin()->second;
+    queue_items.erase(queue_items.begin());
+  }
+
+  // adjust qlen
+  dq->qlen.sub(in_qlen);
+
+  for (map<int,list<Message*> >::iterator p = in_q.begin(); p != in_q.end(); p++)
+    for (list<Message*>::iterator r = p->second.begin(); r != p->second.end(); r++) {
+      if (*r < (void *) DispatchQueue::D_NUM_CODES) {
+        continue; // skip non-Message dispatch codes
+      }
+      msgr->dispatch_throttle_release((*r)->get_dispatch_throttle_size());
+      ldout(msgr->cct,20) << "  discard " << *r << dendl;
+      (*r)->put();
+    }
+  in_q.clear();
+  in_qlen = 0;
+}
+
+void SimpleMessenger::IncomingQueue::restart_queue()
+{
+  halt = false;
+}
+
+
+
 /********************************************
  * SimpleMessenger
  */
@@ -2362,7 +2330,7 @@ void SimpleMessenger::reaper()
       ::close(p->sd);
     ldout(cct,10) << "reaper reaped pipe " << p << " " << p->get_peer_addr() << dendl;
     if (p->connection_state)
-      p->connection_state->clear_pipe();
+      p->connection_state->clear_pipe(p);
     p->put();
     ldout(cct,10) << "reaper deleted pipe " << p << dendl;
   }
@@ -2414,7 +2382,7 @@ int SimpleMessenger::start()
   started = true;
 
   if (!did_bind)
-    ms_addr.nonce = nonce;
+    my_inst.addr.nonce = nonce;
 
   lock.Unlock();
 
@@ -2430,15 +2398,17 @@ int SimpleMessenger::start()
 /* connect_rank
  * NOTE: assumes messenger.lock held.
  */
-SimpleMessenger::Pipe *SimpleMessenger::connect_rank(const entity_addr_t& addr, int type)
+SimpleMessenger::Pipe *SimpleMessenger::connect_rank(const entity_addr_t& addr,
+                                                     int type,
+                                                     Connection *con)
 {
   assert(lock.is_locked());
-  assert(addr != ms_addr);
+  assert(addr != my_inst.addr);
   
   ldout(cct,10) << "connect_rank to " << addr << ", creating pipe and registering" << dendl;
   
   // create pipe
-  Pipe *pipe = new Pipe(this, Pipe::STATE_CONNECTING);
+  Pipe *pipe = new Pipe(this, Pipe::STATE_CONNECTING, con);
   pipe->pipe_lock.Lock();
   pipe->set_peer_type(type);
   pipe->set_peer_addr(addr);
@@ -2468,33 +2438,11 @@ bool SimpleMessenger::verify_authorizer(Connection *con, int peer_type,
   return ms_deliver_verify_authorizer(con, peer_type, protocol, authorizer, authorizer_reply, isvalid);
 }
 
-void SimpleMessenger::submit_message(Message *m, Pipe *pipe)
-{ 
-  lock.Lock();
-  if (pipe == dispatch_queue.local_pipe) {
-    ldout(cct,20) << "submit_message " << *m << " local" << dendl;
-    dispatch_queue.local_delivery(m, m->get_priority());
-  } else {
-    pipe->pipe_lock.Lock();
-    if (pipe->state == Pipe::STATE_CLOSED) {
-      ldout(cct,20) << "submit_message " << *m << " ignoring closed pipe " << pipe->peer_addr << dendl;
-      pipe->unregister_pipe();
-      pipe->pipe_lock.Unlock();
-      m->put();
-    } else {
-      ldout(cct,20) << "submit_message " << *m << " remote " << pipe->peer_addr << dendl;
-      pipe->_send(m);
-      pipe->pipe_lock.Unlock();
-    }
-  }
-  lock.Unlock();
-}
-
 Connection *SimpleMessenger::get_connection(const entity_inst_t& dest)
 {
   Mutex::Locker l(lock);
   Pipe *pipe = NULL;
-  if (ms_addr == dest.addr) {
+  if (my_inst.addr == dest.addr) {
     // local
     pipe = dispatch_queue.local_pipe;
   } else {
@@ -2512,94 +2460,84 @@ Connection *SimpleMessenger::get_connection(const entity_inst_t& dest)
       }
     }
     if (!pipe) {
-      Policy& policy = get_policy(dest.name.type());
-      if (policy.lossy && policy.server)
-	pipe = NULL;
-      else
-	pipe = connect_rank(dest.addr, dest.name.type());
+      pipe = connect_rank(dest.addr, dest.name.type(), NULL);
     }
   }
-  if (pipe)
-    return (Connection *)pipe->connection_state->get();
-  return NULL;
+  return (Connection *)pipe->connection_state->get();
 }
 
 
-void SimpleMessenger::submit_message(Message *m, const entity_addr_t& dest_addr, int dest_type, bool lazy)
+void SimpleMessenger::submit_message(Message *m, Connection *con, const entity_addr_t& dest_addr, int dest_type, bool lazy)
 {
-  // this is just to make sure that a changeset is working properly;
-  // if you start using the refcounting more and have multiple people
-  // hanging on to a message, ditch the assert!
-  assert(m->nref.read() == 1);
-
-  if (dest_addr == entity_addr_t()) {
-    ldout(cct,0) << "submit_message message " << *m << " with empty dest " << dest_addr << dendl;
-    m->put();
-    return;
+  Pipe *pipe = NULL;
+  if (con) {
+    pipe = con ? (Pipe *)con->pipe : NULL;
+    // we don't want to deal with ref-counting here, so we don't use get_pipe()
+    con->get();
   }
 
-  lock.Lock();
-  {
-    // local?
-    if (ms_addr == dest_addr) {
-      if (!destination_stopped) {
-        // local
-        ldout(cct,20) << "submit_message " << *m << " local" << dendl;
-	dispatch_queue.local_delivery(m, m->get_priority());
-      } else {
-        ldout(cct,0) << "submit_message " << *m << " " << dest_addr << " local but no local endpoint, dropping." << dendl;
-        assert(0);  // hmpf, this is probably mds->mon beacon from newsyn.
-        m->put();
-      }
+  // local?
+  if (!pipe && my_inst.addr == dest_addr) {
+    if (!destination_stopped) {
+      // local
+      ldout(cct,20) << "submit_message " << *m << " local" << dendl;
+      dispatch_queue.local_delivery(m, m->get_priority());
     } else {
-      // remote pipe.
-      Pipe *pipe = 0;
-      if (rank_pipe.count(dest_addr)) {
-	pipe = rank_pipe[ dest_addr ];
-	pipe->pipe_lock.Lock();
-	if (pipe->state == Pipe::STATE_CLOSED) {
-	  ldout(cct,20) << "submit_message " << *m << " remote, " << dest_addr << ", ignoring closed pipe." << dendl;
-	  pipe->unregister_pipe();
-	  pipe->pipe_lock.Unlock();
-	  pipe = 0;
-	} else {
-	  ldout(cct,20) << "submit_message " << *m << " remote, " << dest_addr << ", have pipe." << dendl;
-	  
-	  pipe->_send(m);
-	  pipe->pipe_lock.Unlock();
-	}
+      ldout(cct,0) << "submit_message " << *m << " " << dest_addr << " local but no local endpoint, dropping." << dendl;
+      assert(0);  // hmpf, this is probably mds->mon beacon from newsyn.
+      m->put();
+    }
+  } else {
+    // remote pipe.
+    if (pipe) {
+      pipe->pipe_lock.Lock();
+      if (pipe->state == Pipe::STATE_CLOSED) {
+        ldout(cct,0) << "submit_message " << *m << " remote, " << dest_addr << ", ignoring closed pipe, dropping message " << m << dendl;
+        pipe->unregister_pipe();
+        pipe->pipe_lock.Unlock();
+        pipe = 0;
+	assert(con);
+	con->put();
+	return;
+      } else {
+        ldout(cct,20) << "submit_message " << *m << " remote, " << dest_addr << ", have pipe." << dendl;
+
+        pipe->_send(m);
+        pipe->pipe_lock.Unlock();
       }
-      if (!pipe) {
-	Policy& policy = get_policy(dest_type);
-	if (policy.lossy && policy.server) {
-	  ldout(cct,20) << "submit_message " << *m << " remote, " << dest_addr << ", lossy server for target type "
-		   << ceph_entity_type_name(dest_type) << ", no session, dropping." << dendl;
-	  m->put();
-	} else if (lazy) {
-	  ldout(cct,20) << "submit_message " << *m << " remote, " << dest_addr << ", lazy, dropping." << dendl;
-	  m->put();
-	} else {
-	  ldout(cct,20) << "submit_message " << *m << " remote, " << dest_addr << ", new pipe." << dendl;
-	  // not connected.
-	  pipe = connect_rank(dest_addr, dest_type);
-	  pipe->send(m);
-	}
+    }
+    if (!pipe) {
+      const Policy& policy = get_policy(dest_type);
+      if (policy.server) {
+        ldout(cct,20) << "submit_message " << *m << " remote, " << dest_addr << ", lossy server for target type "
+            << ceph_entity_type_name(dest_type) << ", no session, dropping." << dendl;
+        m->put();
+      } else if (lazy) {
+        ldout(cct,20) << "submit_message " << *m << " remote, " << dest_addr << ", lazy, dropping." << dendl;
+        m->put();
+      } else {
+        ldout(cct,20) << "submit_message " << *m << " remote, " << dest_addr << ", new pipe." << dendl;
+        // not connected.
+        pipe = connect_rank(dest_addr, dest_type, con);
+        pipe->send(m);
       }
     }
   }
-
-  lock.Unlock();
+  if (con) {
+    con->put();
+  }
 }
 
 int SimpleMessenger::send_keepalive(const entity_inst_t& dest)
 {
   const entity_addr_t dest_addr = dest.addr;
   entity_addr_t dest_proc_addr = dest_addr;
+  int ret = 0;
 
   lock.Lock();
   {
     // local?
-    if (ms_addr != dest_addr) {
+    if (my_inst.addr != dest_addr) {
       // remote.
       Pipe *pipe = 0;
       if (rank_pipe.count( dest_proc_addr )) {
@@ -2611,33 +2549,40 @@ int SimpleMessenger::send_keepalive(const entity_inst_t& dest)
 	  pipe->unregister_pipe();
 	  pipe->pipe_lock.Unlock();
 	  pipe = 0;
+	  ret = -EPIPE;
 	} else {
 	  ldout(cct,20) << "send_keepalive remote, " << dest_addr << ", have pipe." << dendl;
 	  pipe->_send_keepalive();
 	  pipe->pipe_lock.Unlock();
 	}
+      } else {
+        ret = -EINVAL;
       }
-      if (!pipe)
+      if (!pipe) {
 	ldout(cct,20) << "send_keepalive no pipe for " << dest_addr << ", doing nothing." << dendl;
+      }
     }
   }
   lock.Unlock();
-  return 0;
+  return ret;
 }
 
 int SimpleMessenger::send_keepalive(Connection *con)
 {
+  int ret = 0;
   SimpleMessenger::Pipe *pipe = (SimpleMessenger::Pipe *)con->get_pipe();
   if (pipe) {
     ldout(cct,20) << "send_keepalive con " << con << ", have pipe." << dendl;
+    assert(pipe->msgr == this);
     pipe->pipe_lock.Lock();
     pipe->_send_keepalive();
     pipe->pipe_lock.Unlock();
     pipe->put();
   } else {
     ldout(cct,0) << "send_keepalive con " << con << ", no pipe." << dendl;
+    ret = -EPIPE;
   }
-  return 0;
+  return ret;
 }
 
 
@@ -2751,6 +2696,7 @@ void SimpleMessenger::mark_down(Connection *con)
   Pipe *p = (Pipe *)con->get_pipe();
   if (p) {
     ldout(cct,1) << "mark_down " << con << " -- " << p << dendl;
+    assert(p->msgr == this);
     p->unregister_pipe();
     p->pipe_lock.Lock();
     p->stop();
@@ -2767,6 +2713,7 @@ void SimpleMessenger::mark_down_on_empty(Connection *con)
   lock.Lock();
   Pipe *p = (Pipe *)con->get_pipe();
   if (p) {
+    assert(p->msgr == this);
     p->pipe_lock.Lock();
     p->unregister_pipe();
     if (p->out_q.empty()) {
@@ -2790,9 +2737,9 @@ void SimpleMessenger::mark_disposable(Connection *con)
   Pipe *p = (Pipe *)con->get_pipe();
   if (p) {
     ldout(cct,1) << "mark_disposable " << con << " -- " << p << dendl;
+    assert(p->msgr == this);
     p->pipe_lock.Lock();
     p->policy.lossy = true;
-    p->disposable = true;
     p->pipe_lock.Unlock();
     p->put();
   } else {
@@ -2804,13 +2751,13 @@ void SimpleMessenger::mark_disposable(Connection *con)
 void SimpleMessenger::learned_addr(const entity_addr_t &peer_addr_for_me)
 {
   // be careful here: multiple threads may block here, and readers of
-  // ms_addr do NOT hold any lock.
+  // my_inst.addr do NOT hold any lock.
   lock.Lock();
   if (need_addr) {
     entity_addr_t t = peer_addr_for_me;
-    t.set_port(ms_addr.get_port());
-    ms_addr.addr = t.addr;
-    ldout(cct,1) << "learned my addr " << ms_addr << dendl;
+    t.set_port(my_inst.addr.get_port());
+    my_inst.addr.addr = t.addr;
+    ldout(cct,1) << "learned my addr " << my_inst.addr << dendl;
     need_addr = false;
     init_local_pipe();
   }
@@ -2819,6 +2766,6 @@ void SimpleMessenger::learned_addr(const entity_addr_t &peer_addr_for_me)
 
 void SimpleMessenger::init_local_pipe()
 {
-  dispatch_queue.local_pipe->connection_state->peer_addr = msgr->ms_addr;
+  dispatch_queue.local_pipe->connection_state->peer_addr = msgr->my_inst.addr;
   dispatch_queue.local_pipe->connection_state->peer_type = msgr->my_type;
 }
