@@ -246,7 +246,7 @@ void Objecter::shutdown()
   }
 }
 
-void Objecter::send_linger(LingerOp *info)
+void Objecter::send_linger(LingerOp *info, bool first_send)
 {
   if (!info->registering) {
     ldout(cct, 15) << "send_linger " << info->linger_id << dendl;
@@ -264,7 +264,20 @@ void Objecter::send_linger(LingerOp *info)
 	linger_check_for_latest_map(info);
       }
     }
-    op_submit(o, info->session);
+
+    if (first_send) {
+      op_submit(o);
+    } else {
+      _op_submit(o);
+    }
+
+    OSDSession *s = o->session;
+    if (info->session != s) {
+      info->session_item.remove_myself();
+      info->session = s;
+      if (info->session)
+	s->linger_ops.push_back(&info->session_item);
+    }
     info->registering = true;
 
     logger->inc(l_osdc_linger_send);
@@ -335,7 +348,7 @@ tid_t Objecter::linger(const object_t& oid, const object_locator_t& oloc,
 
   logger->set(l_osdc_linger_active, linger_ops.size());
 
-  send_linger(info);
+  send_linger(info, true);
 
   return info->linger_id;
 }
@@ -537,7 +550,7 @@ void Objecter::handle_osd_map(MOSDMap *m)
     LingerOp *op = *p;
     if (op->session) {
       logger->inc(l_osdc_linger_resend);
-      send_linger(op);
+      send_linger(op, false);
     }
   }
 
@@ -742,7 +755,7 @@ void Objecter::kick_requests(OSDSession *session)
   // resend lingers
   for (xlist<LingerOp*>::iterator j = session->linger_ops.begin(); !j.end(); ++j) {
     logger->inc(l_osdc_linger_resend);
-    send_linger(*j);
+    send_linger(*j, false);
   }
 }
 
@@ -780,6 +793,17 @@ void Objecter::tick()
       ++laggy_ops;
     }
   }
+  for (map<uint64_t,LingerOp*>::iterator p = linger_ops.begin();
+       p != linger_ops.end();
+       p++) {
+    LingerOp *op = p->second;
+    if (op->session) {
+      ldout(cct, 10) << " pinging osd that serves lingering tid " << p->first << " (osd." << op->session->osd << ")" << dendl;
+      toping.insert(op->session);
+    } else {
+      ldout(cct, 10) << " lingering tid " << p->first << " does not have session" << dendl;
+    }
+  }
   logger->set(l_osdc_op_laggy, laggy_ops);
   logger->set(l_osdc_osd_laggy, toping.size());
 
@@ -791,8 +815,9 @@ void Objecter::tick()
     // (osd reply message policy is lossy)
     for (set<OSDSession*>::iterator i = toping.begin();
 	 i != toping.end();
-	 i++)
+	 i++) {
       messenger->send_message(new MPing, (*i)->con);
+    }
   }
     
   // reschedule
@@ -831,7 +856,7 @@ void Objecter::resend_mon_ops()
 
 // read | write ---------------------------
 
-tid_t Objecter::op_submit(Op *op, OSDSession *s)
+tid_t Objecter::op_submit(Op *op)
 {
   assert(client_lock.is_locked());
   assert(initialized);
@@ -844,6 +869,11 @@ tid_t Objecter::op_submit(Op *op, OSDSession *s)
   // take_op_budget() may drop our lock while it blocks.
   take_op_budget(op);
 
+  return _op_submit(op);
+}
+
+tid_t Objecter::_op_submit(Op *op)
+{
   // pick tid
   tid_t mytid = ++last_tid;
   op->tid = mytid;
@@ -851,14 +881,9 @@ tid_t Objecter::op_submit(Op *op, OSDSession *s)
 
   // pick target
   bool check_for_latest_map = false;
-  if (s) {
-    op->session = s;
-    s->ops.push_back(&op->session_item);
-  } else {
-    int r = recalc_op_target(op);
-    check_for_latest_map = (r == RECALC_OP_TARGET_POOL_DNE);
-    num_homeless_ops++;  // initially!
-  }
+  num_homeless_ops++;  // initially; recalc_op_target() will decrement if it finds a target
+  int r = recalc_op_target(op);
+  check_for_latest_map = (r == RECALC_OP_TARGET_POOL_DNE);
 
   // add to gather set(s)
   if (op->onack) {
@@ -1269,7 +1294,8 @@ void Objecter::handle_osd_op_reply(MOSDOpReply *m)
   if (!op->onack && !op->oncommit) {
     op->session_item.remove_myself();
     ldout(cct, 15) << "handle_osd_op_reply completed tid " << tid << dendl;
-    put_op_budget(op);
+    if (op->budgeted)
+      put_op_budget(op);
     ops.erase(tid);
     logger->set(l_osdc_op_active, ops.size());
     if (op->con)
@@ -1354,7 +1380,8 @@ void Objecter::list_objects(ListContext *list_context, Context *onfinish) {
   op_submit(o);
 }
 
-void Objecter::_list_reply(ListContext *list_context, bufferlist *bl, Context *final_finish, epoch_t reply_epoch)
+void Objecter::_list_reply(ListContext *list_context, int r, bufferlist *bl,
+			   Context *final_finish, epoch_t reply_epoch)
 {
   ldout(cct, 10) << "_list_reply" << dendl;
 
@@ -1379,14 +1406,24 @@ void Objecter::_list_reply(ListContext *list_context, bufferlist *bl, Context *f
   if (response_size) {
     ldout(cct, 20) << "got a response with objects, proceeding" << dendl;
     list_context->list.merge(response.entries);
-    list_context->max_entries -= response_size;
-    ldout(cct, 20) << "cleaning up and exiting" << dendl;
-    if (!list_context->max_entries) {
+    if (response_size >= list_context->max_entries) {
       final_finish->finish(0);
       delete bl;
       delete final_finish;
       return;
     }
+
+    // ask for fewer objects next time around
+    list_context->max_entries -= response_size;
+
+    // if the osd returns 1 (newer code), or no entries, it means we
+    // hit the end of the pg.
+    if (r == 0 && response_size > 0) {
+      // not yet done with this pg
+      delete bl;
+      list_objects(list_context, final_finish);
+      return;
+    }
   }
 
   // if we make this this far, there are no objects left in the current pg, but we want more!
@@ -2016,7 +2053,7 @@ Objecter::RequestStateHook::RequestStateHook(Objecter *objecter) :
 {
 }
 
-bool Objecter::RequestStateHook::call(std::string command, bufferlist& out)
+bool Objecter::RequestStateHook::call(std::string command, std::string args, bufferlist& out)
 {
   stringstream ss;
   JSONFormatter formatter(true);
